unit im_Histogram;

interface

uses
  windows,graphics,GR32,GR32_Image,im_basic;

  TYPE
    THistoArray  = ARRAY[BYTE] OF INTEGER;

    TColorPlane  = (cpRed, cpGreen, cpBlue,   // RGB color space
                    cpValue,                  // V from HSV color space
                    cpIntensity);             // Intensity "plane"

    TRGBHistoArray =
      RECORD
        Red      :  THistoArray;
        Green    :  THistoArray;
        Blue     :  THistoArray;
        Intensity:  THistoArray;
      END;

    THistogram =
      CLASS(TObject)
        PRIVATE
          FHistogram:  THistoArray;
          FUNCTION GetCount:  INTEGER;
        PUBLIC
          CONSTRUCTOR Create;
          PROCEDURE Draw (CONST Canvas:  TCanvas);
          PROCEDURE Clear;
          PROCEDURE Increment(CONST index:  BYTE);
          PROPERTY Frequency:  THistoArray READ FHistogram WRITE FHistogram;
          PROPERTY Count    :  INTEGER     READ GetCount;
end;
          PROCEDURE GetHistogram(CONST ColorPlane:  TColorPlane;
                         CONST Bitmap:  TBitmap32;
                         VAR   Histogram:  THistogram);


IMPLEMENTATION

  USES
    Dialogs,                    // ShowMessage
    Math,                       // MaxIntValue, IntPwr
    SysUtils;                   // pByteArray, Exception


{$R *.dfm}

  CONSTRUCTOR THistogram.Create;
  BEGIN
    Clear
  END {Create};

  FUNCTION THistogram.GetCount:  INTEGER;
    VAR
      i:  BYTE;
  BEGIN
    RESULT := 0;
    FOR i := Low(THistoArray) TO High(THistoArray) DO
      INC(RESULT, FHistogram[i])
  END {GetCount};

PROCEDURE THistogram.Clear;
    VAR
      i:  BYTE;
  BEGIN
    FOR i :=  Low(THistoArray) TO High(THistoArray) DO
      FHistogram[i] := 0
END {Clear};

PROCEDURE THistogram.Increment(CONST index:  BYTE);
  BEGIN
    INC(FHistogram[index])
  END {Increment};


  PROCEDURE GetHistogram(CONST ColorPlane:  TColorPlane;
                         CONST Bitmap:  TBitmap32;
                         VAR   Histogram:  THistogram);
    VAR
      i    :  INTEGER;
      index:  INTEGER;
      j    :  INTEGER;
    row:Pcolor32array;
     h1,h2,h3:integer;
     rd,gr,bl:integer;
BEGIN
    h1:=$0000FF; //0
    h2:=$00FF00; //1
    h3:=$FF0000; //2

   histogram.Clear;
    // Step through each row of image.

    FOR j := Bitmap.Height-1 DOWNTO 0 DO
    BEGIN
      Row  := Bitmap.Scanline[j];

      FOR i := Bitmap.Width-1 DOWNTO 0 DO
      BEGIN
      Rd  :=row[i] and h1;
      Gr:=(row[i] and h2) shr 8;
      Bl:=(row[i] and h3) shr 16;

      CASE ColorPlane OF
      cpRed   :  index:=rd;
      cpGreen :  index:=gr;
      cpBlue  :   index:=bl;
      // V from HSV Color Space
      cpValue :  index := MaxIntValue( [rd,gr,bl] );
      cpIntensity: index := (rd + gr + bl) DIV 3
        ELSE
            index := 0;   // should never happen; avoid compiler warning
        END;
        histogram.Increment(index)
      END

    END

  END {GetHistogram};

  // Draw Histogram on specified Canvas, assumed to have
  // a width that is a multiple of 256 (for best results).




PROCEDURE thistogram.Draw(CONST Canvas:  Tcanvas);
    CONST
      MajorTickSize = 8;

    VAR
      BarLength:  INTEGER;
      Delta    :  INTEGER;
      Color    :  TColor;
      factor   :  DOUBLE;
      i        :  INTEGER;
      index    :  INTEGER;
      j        :  INTEGER;
      Height   :  INTEGER;
      MaxValue :  INTEGER;
      Width    :  INTEGER;
  BEGIN
    Height   := Canvas.ClipRect.Bottom;
    Width    := Canvas.ClipRect.Right;

    MaxValue := MaxIntValue(SELF.Frequency);
    factor := Width / 256;

    // For now only paint on a canvas exactly 256 pixels wide.  If
    // MaxValue is zero, array was not filled in correctly and is ignored.
    IF   MaxValue > 0
    THEN BEGIN
      FOR i := 0 TO Width-1 DO
      BEGIN
        // In general, Width should be multiple of 255 + 1
        index := ROUND(i / factor);
        index := MinIntValue( [255, index] );

        Color := RGB(index, index, index);
        Canvas.Pen.Color := Color;
        BarLength := ROUND(Height*SELF.Frequency[index] / MaxValue);
        Canvas.MoveTo(i, Height-1);
        Canvas.LineTo(i, Height-1-BarLength)
      END;

      Canvas.Pen.Color := clRed;
      // Vertical Lines for visual estimation
      FOR i := 0 TO 25 DO
      BEGIN
        Canvas.MoveTo(Round(10*i*factor), Height-1);
        IF   i MOD 5 = 0
        THEN Delta := MajorTickSize
        ELSE Delta := MajorTickSize DIV 2;
        Canvas.LineTo(Round(10*i*factor), Height-1-Delta);
      END;

      // Horizontal Lines
      FOR j := 1 TO 4 DO
      BEGIN
        Canvas.MoveTo(      0, j*Height DIV 5);
        Canvas.LineTo(Width-1, j*Height DIV 5);
      END;

      Canvas.Brush.Style := bsClear;
      Canvas.Font.Color := clRed;
      Canvas.TextOut(2,2, 'Max = ' + IntToStr(MaxValue));
      Canvas.TextOut(2, Height-Canvas.TextHeight('X') - MajorTickSize, '0');
      Canvas.TextOut(Width-Canvas.TextWidth('250 '),
                     Height-Canvas.TextHeight('X') - MajorTickSize, '250')
    END
  END {DrawHistogram};

end.
